

from typing import Callable, Iterable, List, Dict
from pyparsing import Optional

import torch
import os
from functools import partial
from collections import defaultdict


from .dict2obj import Config
from .utils import AverageMeter, getLogger
from .metrics import *


DEFAULT_METRICS = {
    'LOSS': None,
    'NDCG': normalized_dcg,
    'RECALL': recall,
    'PRECISION': precision,
    'HITRATE': hit_rate,
    'MSE': mean_squared_error,
    'MAE': mean_abs_error,
    'RMSE': root_mse
}

DEFAULT_FMTS = {
    'LOSS': ".5f",
    'NDCG': ".4f",
    'RECALL': ".4f",
    'PRECISION': ".3%",
    'HITRATE': ".3%",
    'MSE': ".4f",
    'MAE': ".4f",
    'RMSE': ".4f"
}


class Coach:
    
    def __init__(
        self, model: torch.nn.Module,
        criterion: Callable, 
        optimizer: torch.optim.Optimizer, 
        lr_scheduler,
        device: torch.device
    ):
        self.model = model
        self.device = device
        self.criterion = criterion
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

        self._best = 0.
        self.steps = 0

      
    def save(self) -> None:
        torch.save(self.model.state_dict(), os.path.join(self.cfg.INFO_PATH, self.cfg.FILENAME))

    def save_checkpoint(self, epoch: int) -> None:
        path = os.path.join(self.cfg.INFO_PATH, self.cfg.CHECKPOINT_FILENAME)
        checkpoint = dict()
        checkpoint['epoch'] = epoch
        for module in self.CHECKPOINT_MODULES:
            checkpoint[module] = getattr(self, module).state_dict()
        torch.save(checkpoint, path)

    def load_checkpoint(self) -> int:
        path = os.path.join(self.cfg.INFO_PATH, self.cfg.CHECKPOINT_FILENAME)
        checkpoint = torch.load(path)
        for module in self.CHECKPOINT_MODULES:
            getattr(self, module).load_state_dict(checkpoint[module])
        return checkpoint['epoch']

    def save_best(self, path: str, prefix: str): ...

    def check_best(self, results: dict): ...
    
    def compile(
        self, cfg: Config, callbacks: List[str]
    ):
        self.cfg = cfg
        self.meters = Config(
            train=defaultdict(list),
            valid=defaultdict(list)
        )
        for name in callbacks:
            if '@' in callback:
                callback, K = name.split('@')
                for prefix in self.meters:
                    self.meters[prefix][callback].append(
                        AverageMeter(
                            name=name,
                            metric=partial(DEFAULT_METRICS[callback], k=int(K)),
                            fmt=DEFAULT_FMTS[callback]
                        )
                    )
            else:
                callback = name
                for prefix in self.meters:
                    self.meters[prefix][callback].append(
                        AverageMeter(
                            name=name,
                            metric=DEFAULT_METRICS[callback],
                            fmt=DEFAULT_FMTS[callback]
                        )
                    )

    def callback(
        self, *values,
        n: int = 1, mode: str = 'mean', 
        prefix: str = 'train', pool: Optional[List] = None
    ):
        metrics: Dict[List] = self.meters[prefix]
        for metric in pool:
            for meter in metrics.get(metric, []):
                meter(*values, n=n, mode=mode)

    def step(self, epoch: int):
        for prefix, callbacks in self.meters.items():
            callbacks: defaultdict[str, List[AverageMeter]]
            infos = [f"[Epoch: {epoch:<4d}]" + prefix + " >>> "]
            for meters in callbacks.values():
                infos += [meter.step() for meter in meters if meter.active]
            getLogger().info('\t'.join(infos))

    def summary(self):
        for prefix, callbacks in self.meters.items():
            callbacks: defaultdict[str, List[AverageMeter]]
            for meters in callbacks.values():
                for meter in meters:
                    meter.plot()
                    meter.save(path=self.cfg.LOG_PATH, prefix=prefix)

    def train(self, trainloader: Iterable):
        self.model.train()
        inputs: Dict[str, torch.Tensor]
        targets: torch.Tensor
        for inputs, targets in trainloader:
            inputs = {field: val.to(self.device) for field, val in inputs.items()}
            targets = targets.to(self.device)

            outs = self.model(inputs)
            loss = self.criterion(outs, targets)

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            
            self.callback(loss.item(), n=targets.size(0), mode="mean", prefix='train', pool=['LOSS'])

        self.lr_scheduler.step() # TODO: step() per epoch or per mini-batch ?


    @torch.no_grad()
    def evaluate(self, dataloader: Iterable, prefix: str = 'valid'):
        self.model.eval()
        inputs: Dict[str, torch.Tensor]
        targets: Dict[str, torch.Tensor]
        for inputs, targets in dataloader:
            inputs = {field: val.to(self.device) for field, val in inputs.items()}
            targets = {field: val.to(self.device) for field, val in targets.items()}

            preds = self.model(inputs)
            loss = self.criterion(preds, targets)

            self.callback(
                preds, targets, 
                n=targets.size(0), mode="mean", prefix=prefix,
                pool=['NDCG', 'RECALL', 'PRECISION', 'HITRATE', 'MSE', 'MAE', 'RMSE']
            )
            if prefix == 'valid':
                self.callback(loss, n=targets.size(0), mode="mean", prefix=prefix, pool=['LOSS'])


    def fit(
        self, trainloader: Iterable, validloader: Iterable,
        *, epochs: int, start_epoch: int = 0
    ):
        for epoch in range(start_epoch, epochs):
            if epoch % self.cfg.CHECKPOINT_FREQ == 0:
                self.save_checkpoint(epoch)
            if epoch % self.cfg.EVAL_FREQ == 0:
                if self.cfg.EVAL_TRAIN:
                    self.evaluate(trainloader, prefix='train')
                if self.cfg.EVAL_VALID:
                    results = self.evaluate(validloader, prefix='valid')
                    self.check_best(results)
            
            self.train(trainloader)

            self.step(epoch)
        self.save()
        self.summary()






from typing import Callable, Optional, Any
import torch
import numpy as np

import logging
import time
import random
import os
from freeplot.base import FreePlot

from .dict2obj import Config


LOGGER = Config(
    name='RecSys', filename='log.txt', level=logging.DEBUG,
    filelevel=logging.DEBUG, consolelevel=logging.INFO,
    formatter=Config(
        filehandler=logging.Formatter('%(asctime)s:\t%(message)s'),
        consolehandler=logging.Formatter('%(message)s')
    )
)


def _unitary(value):
    return value

class AverageMeter:

    def __init__(self, name: str, metric: Optional[Callable] = None, fmt: str = ".5f"):
        self.name = name
        self.fmt = fmt
        self.reset()
        self.history = []
        self.active = False
        self.__metric = metric if metric else _unitary

    def reset(self) -> None:
        self.val = 0.
        self.avg = 0.
        self.sum = 0.
        self.count = 0

    def update(self, val: float, n: int = 1, mode: str = "mean") -> None:
        self.val = val
        self.count += n
        if mode == "mean":
            self.sum += val * n
        elif mode == "sum":
            self.sum += val
        else:
            raise ValueError(f"Receive mode {mode} but 'mean' or 'sum' expected ...")
        self.avg = self.sum / self.count

    def step(self) -> str:
        self.history.append(self.avg)
        self.reset()
        return str(self)

    def callback(self, *values):
        val = self.__metric(*values)
        try:
            return val.item() # Tensor|ndarray
        except AttributeError:
            return val # float

    def plot(self) -> None:
        self.fp = FreePlot(
            shape=(1, 1),
            figsize=(2.2, 2),
            titles=(self.name,),
            dpi=300, latex=False
        )
        timeline = np.arange(1, len(self.history) + 1)
        self.fp.lineplot(timeline, self.history, marker='')
        self.fp.set_title(y=.98)
    
    def save(self, path: str, prefix: str = '') -> None:
        filename = f"{prefix}{self.name}.png"
        self.fp.savefig(os.path.join(path, filename))

    def __str__(self):
        fmtstr = "{name} Avg:{avg:{fmt}}"
        return fmtstr.format(**self.__dict__)

    def __call__(self, *values, n: int = 1, mode: str = "mean")  -> None:
        self.active = True
        self.update(
            val = self.callback(*values),
            n = n,
            mode = mode
        )


def set_logger(
    path: str,
    log2file: bool = True, log2console: bool = True
) -> None:
    logger = logging.getLogger(LOGGER.name)
    logger.setLevel(LOGGER.level)

    if log2file:
        handler = logging.FileHandler(
            os.path.join(path, LOGGER.filename), 
            encoding='utf-8'
        )
        handler.setLevel(LOGGER.filelevel)
        handler.setFormatter(LOGGER.formatter.filehandler)
        logger.addHandler(handler)
    if log2console:
        handler = logging.StreamHandler()
        handler.setLevel(LOGGER.consolelevel)
        handler.setFormatter(LOGGER.formatter.consolehandler)
        logger.addHandler(handler)
    logger.debug("========================================================================")
    logger.debug("========================================================================")
    logger.debug("========================================================================")
    return logger

def getLogger():
    return logging.getLogger(LOGGER.name)

def timemeter(prefix=""):
    def decorator(func):
        logger = getLogger()
        def wrapper(*args, **kwargs):
            start = time.time()
            results = func(*args, **kwargs)
            end = time.time()
            logger.info(f"[Wall TIME]- {prefix} takes {end-start:.6f} seconds ...")
            return  results
        wrapper.__doc__ = func.__doc__
        wrapper.__name__ = func.__name__
        return wrapper
    return decorator

def mkdirs(*paths: str) -> None:
    for path in paths:
        try:
            os.makedirs(path)
        except FileExistsError:
            pass


def activate_benchmark(benchmark: bool) -> None:
    from torch.backends import cudnn
    if benchmark:
        getLogger().info(f"[Seed] >>> Activate benchmark")
        cudnn.benchmark, cudnn.deterministic = True, False
    else:
        getLogger().info(f"[Seed] >>> Deactivate benchmark")
        cudnn.benchmark, cudnn.deterministic = False, True

def set_seed(seed: int) -> None:
    if seed == -1:
        seed = random.randint(0, 1024)
        logger = getLogger()
        logger.info(f"[Seed] >>> Set seed randomly: {seed}")
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)








import torch
import torch.nn as nn
from torchrec.modules.deepfm import FactorizationMachine


__all__ = ['FM']

class FM(FactorizationMachine):
    """
    1. List[Tensor: B x k x d] -> Tensor: B x [k1 x d1 + k2 x d2 ...] -> Tensor: B x 1
    2. List[Tensor: B x d] -> Tensor: B x [d1 + d2 ...] -> Tensor: B x 1
    """







import torch
import time
import yaml
from argparse import ArgumentParser

from .dict2obj import Config
from .utils import mkdirs, set_logger, set_seed, activate_benchmark, timemeter


INFO_PATH = "./infos/{description}"
LOG_PATH = "./logs/{description}-{TIME}"
TIME = "%m%d%H"

CONFIG = Config(
    DEVICE = "cuda" if torch.cuda.is_available() else "cpu",
    BENCHMARK = True, # activate cudnn.benchmark if True
    SEED = -1, # -1 for random

    # evaluation
    EVAL_FREQ = 5,
    EVAL_TRAIN = False,
    EVAL_VALID = True,

    # logger
    log2file = True, 
    log2console = True,

    # path|file
    SAVED_FILENAME = "model.pt", # the filename of saved model paramters
    CHECKPOINT_FREQ = 1,
    CHECKPOINT_MODULES = ['model', 'optimizer', 'lr_scheduler'],
    CHECKPOINT_FILENAME = "checkpoint.tar",
    description = "RecSys"

)


class Parser(Config):

    def __init__(self) -> None:
        super().__init__(**CONFIG)

    def readme(self, path: str, mode: str = "w") -> None:
        time_ = time.strftime("%Y-%m-%d-%H:%M:%S")
        filename = path + "/README.md"
        s = "|  {key}  |   {val}    |\n"
        info = "\n## {0} \n\n\n".format(time_)
        info += "|  Attribute   |   Value   |\n"
        info += "| :-------------: | :-----------: |\n"
        for key, val in self.items():
            info += s.format(key=key, val=val)
        with open(filename, mode, encoding="utf8") as fh:
            fh.write(info)

    def load(self, args: ArgumentParser):
        if hasattr(args, 'config') and args.config:
            with open(args.config, encoding="UTF-8", mode='r') as f:
                for key, val in yaml.full_load(f).items():
                    if key.upper() in self:
                        self[key.upper()] = val
                    else:
                        self[key] = val

    def reset(self):
        self.clear()
        for key, val in CONFIG.items():
            self[key] = val

    @timemeter("Parser/compile")
    def compile(self, args: ArgumentParser):
        for key, val in args._get_kwargs():
            if key.upper() in self:
                self[key.upper()] = val
            else:
                self[key] = val
        self.load(args) # loading config (.yaml) ...
        
        self['TIME'] = time.strftime(TIME)
        self['INFO_PATH'] = INFO_PATH.format(**self)
        self['LOG_PATH'] = LOG_PATH.format(**self)
        mkdirs(self.INFO_PATH, self.LOG_PATH)
        set_logger(path=self.LOG_PATH, log2file=self.log2file, log2console=self.log2console)

        self.DEVICE = torch.device(self.DEVICE)
        activate_benchmark(self.BENCHMARK)
        set_seed(self.SEED)

        self.readme(self.INFO_PATH) # generate README.md
        self.readme(self.LOG_PATH)



import setuptools

with open("README.md", "r") as fh:
  long_description = fh.read()

requires = [
    "tqdm>=4.64.0",
    "freeplot>=0.0.12"
]

setuptools.setup(
  name="freerec",
  version="0.0.1",
  author="MTandHJ",
  author_email="congxueric@gmail.com",
  description="PyTorch library for recommender systems",
  long_description=long_description,
  long_description_content_type="text/markdown",
  license='MIT License',
  url="https://github.com/MTandHJ/freeplot",
  packages=setuptools.find_packages(),
  python_requires='>=3.9',
  install_requires=requires,
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent",
  ],
)





import torch.nn as nn

from ..data.fields import Tokenizer

__all__ = ['RecSysArch']

class RecSysArch(nn.Module):

    def __init__(self, tokenizer: Tokenizer) -> None:
        super().__init__()

        self.tokenizer = tokenizer






from typing import Dict, List

import torch
import torch.nn as nn
import torch.nn.functional as F

from .base import RecSysArch

from ..data.fields import Tokenizer
from ..layers import FM


__all__ = ['DeepFM']

class DeepFM(RecSysArch):

    def __init__(self, tokenizer: Tokenizer) -> None:
        super().__init__(tokenizer)

        sparse_dim = self.tokenizer.dimension('sparse')
        dense_dim = self.tokenizer.dimension('dense')
        self.linear = nn.Linear(dense_dim, 1)
        self.fm = FM()
        self.dnn = nn.Sequential(
            nn.Linear(sparse_dim + dense_dim, 256),
            # nn.BatchNorm1d(256),
            nn.ReLU(True),
            # nn.Dropout(0.)
            nn.Linear(256, 128),
            # nn.BatchNorm1d(128),
            nn.ReLU(True),
            # nn.Dropout(0.)
        )
        self.bias = nn.parameter.Parameter(torch.zeros((1,)), requires_grad=True)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> torch.Tensor:
        sparse: List[torch.Tensor] = self.tokenizer.look_up(inputs)
        dense: torch.Tensor = self.tokenizer.collect_dense(inputs)

        outs_linear = self.linear(dense)
        outs_fm = self.fm(sparse)
        outs_dnn = self.dnn(
            self.tokenizer.flatten_cat(sparse + [dense])
        )
        outs = outs_linear + outs_fm + outs_dnn + self.bias

        return F.sigmoid(outs) # B x 1








"""
Refer to
    https://zhuanlan.zhihu.com/p/67287992
for formal definitions.
"""

from multiprocessing import reduction
from typing import Optional, Union, List 

import torch
import torchmetrics

from functools import partial


__all__ = ['normalized_dcg', 'recall', 'precision', 'hit_rate', 'mean_abs_error', 'mean_squared_error', 'root_mse']

def _reduce(reduction='mean'):
    def decorator(func):
        def wrapper(*args, reduction = reduction, **kwargs):
            results = func(*args, **kwargs)
            if reduction == 'none':
                return results
            elif reduction == 'mean':
                return results.mean()
            elif reduction == 'sum':
                return results.sum()
            else:
                raise ValueError(f"reduction should be 'none'|'mean'|'sum' but {reduction} is received ...")
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        return wrapper
    return decorator


@_reduce('mean')
def normalized_dcg(preds: Union[List[torch.Tensor], torch.Tensor], targets: torch.Tensor, k: Optional[int] = None) -> torch.Tensor:
    """
    Args: each row in Tensor or each element in List denotes a query
        preds: List or Tensor
        target: List or Tensor
    Kwargs:
        k: top-K
        reduction: 'none'|'mean'(default)|'sum'
    """
    if isinstance(preds, torch.Tensor) and preds.ndim == 1:
        preds = preds.unsqueeze(0)
    if isinstance(targets, torch.Tensor) and targets.ndim == 1:
        targets = targets.unsqueeze(0)

    metric = partial(
        torchmetrics.functional.retrieval_normalized_dcg,
        k=k
    )

    return torch.tensor(list(map(metric, preds, targets)))

@_reduce('mean')
def recall(preds: Union[List[torch.Tensor], torch.Tensor], targets: torch.Tensor, k: Optional[int] = None) -> torch.Tensor:
    """
    Args: each row in Tensor or each element in List denotes a query
        preds: List or Tensor
        target: List or Tensor
    Kwargs:
        k: top-K
        reduction: 'none'|'mean'(default)|'sum'
    """
    if isinstance(preds, torch.Tensor) and preds.ndim == 1:
        preds = preds.unsqueeze(0)
    if isinstance(targets, torch.Tensor) and targets.ndim == 1:
        targets = targets.unsqueeze(0)

    metric = partial(
        torchmetrics.functional.retrieval_recall,
        k=k
    )

    return torch.tensor(list(map(metric, preds, targets)))

@_reduce('mean')
def precision(preds: Union[List[torch.Tensor], torch.Tensor], targets: torch.Tensor, k: Optional[int] = None) -> torch.Tensor:
    """
    Args: each row in Tensor or each element in List denotes a query
        preds: List or Tensor
        target: List or Tensor
    Kwargs:
        k: top-K
        reduction: 'none'|'mean'(default)|'sum'
    """
    if isinstance(preds, torch.Tensor) and preds.ndim == 1:
        preds = preds.unsqueeze(0)
    if isinstance(targets, torch.Tensor) and targets.ndim == 1:
        targets = targets.unsqueeze(0)

    metric = partial(
        torchmetrics.functional.retrieval_precision,
        k=k
    )

    return torch.tensor(list(map(metric, preds, targets)))

@_reduce('mean')
def mean_reciprocal_rank(preds: Union[List[torch.Tensor], torch.Tensor], targets: torch.Tensor) -> torch.Tensor:
    """
    Args: each row in Tensor or each element in List denotes a query
        preds: List or Tensor
        target: List or Tensor
    Kwargs:
        reduction: 'none'|'mean'(default)|'sum'
    """
    if isinstance(preds, torch.Tensor) and preds.ndim == 1:
        preds = preds.unsqueeze(0)
    if isinstance(targets, torch.Tensor) and targets.ndim == 1:
        targets = targets.unsqueeze(0)

    metric = torchmetrics.functional.retrieval_reciprocal_rank

    return torch.tensor(list(map(metric, preds, targets)))

@_reduce('mean')
def mean_average_precision(preds: Union[List[torch.Tensor], torch.Tensor], targets: torch.Tensor) -> torch.Tensor:
    """
    Args: each row in Tensor or each element in List denotes a query
        preds: List or Tensor
        target: List or Tensor
    Kwargs:
        reduction: 'none'|'mean'(default)|'sum'
    """
    if isinstance(preds, torch.Tensor) and preds.ndim == 1:
        preds = preds.unsqueeze(0)
    if isinstance(targets, torch.Tensor) and targets.ndim == 1:
        targets = targets.unsqueeze(0)

    metric = torchmetrics.functional.retrieval_average_precision

    return torch.tensor(list(map(metric, preds, targets)))

@_reduce('mean')
def hit_rate(preds: Union[List[torch.Tensor], torch.Tensor], targets: torch.Tensor, k: Optional[int] = None) -> torch.Tensor:
    """
    Args: each row in Tensor or each element in List denotes a query
        preds: List or Tensor
        target: List or Tensor
    Kwargs:
        k: top-K
        reduction: 'mean'(default); formally, the definition of hit rate is set specific
    """
    if isinstance(preds, torch.Tensor) and preds.ndim == 1:
        preds = preds.unsqueeze(0)
    if isinstance(targets, torch.Tensor) and targets.ndim == 1:
        targets = targets.unsqueeze(0)

    metric = partial(
        torchmetrics.functional.retrieval_hit_rate,
        k=k
    )

    return torch.tensor(list(map(metric, preds, targets)))

@_reduce('mean')
def mean_abs_error(preds: Union[List[torch.Tensor], torch.Tensor], targets: torch.Tensor) -> torch.Tensor:
    """
    Args: each row in Tensor or each element in List denotes a query
        preds: List or Tensor
        target: List or Tensor
    Kwargs:
        reduction: 'none'|'mean'(default)|'sum'
    """
    if isinstance(preds, torch.Tensor) and preds.ndim == 1:
        preds = preds.unsqueeze(0)
    if isinstance(targets, torch.Tensor) and targets.ndim == 1:
        targets = targets.unsqueeze(0)

    metric = torchmetrics.functional.mean_absolute_error

    return torch.tensor(list(map(metric, preds, targets)))

@_reduce('mean')
def mean_squared_error(preds: Union[List[torch.Tensor], torch.Tensor], targets: torch.Tensor) -> torch.Tensor:
    """
    Args: each row in Tensor or each element in List denotes a query
        preds: List or Tensor
        target: List or Tensor
    Kwargs:
        reduction: 'none'|'mean'(default)|'sum'
    """
    if isinstance(preds, torch.Tensor) and preds.ndim == 1:
        preds = preds.unsqueeze(0)
    if isinstance(targets, torch.Tensor) and targets.ndim == 1:
        targets = targets.unsqueeze(0)

    metric = partial(
        torchmetrics.functional.mean_squared_error,
        squared=True
    )

    return torch.tensor(list(map(metric, preds, targets)))

@_reduce('mean')
def root_mse(preds: Union[List[torch.Tensor], torch.Tensor], targets: torch.Tensor) -> torch.Tensor:
    """
    Args: each row in Tensor or each element in List denotes a query
        preds: List or Tensor
        target: List or Tensor
    Kwargs:
        reduction: 'none'|'mean'(default)|'sum'
    """
    if isinstance(preds, torch.Tensor) and preds.ndim == 1:
        preds = preds.unsqueeze(0)
    if isinstance(targets, torch.Tensor) and targets.ndim == 1:
        targets = targets.unsqueeze(0)

    metric = partial(
        torchmetrics.functional.mean_squared_error,
        squared=False
    )

    return torch.tensor(list(map(metric, preds, targets)))


if __name__ == "__main__":
    preds = [torch.tensor([0.2, 0.3, 0.5]), torch.tensor([0.1, 0.3, 0.5, 0.2])]
    targets = [torch.tensor([False, False, True]), torch.tensor([False, True, False, True])]
    print(mean_reciprocal_rank(preds, targets))





from typing import Any

# Convert the dict to a object.


class Config(dict):
    '''
    >>> cfg = Config({1:2}, a=3)
    Traceback (most recent call last):
    ...
    TypeError: attribute name must be string, not 'int'
    >>> cfg = Config(a=1, b=2)
    >>> cfg.a
    1
    >>> cfg['a']
    1
    >>> cfg['c'] = 3
    >>> cfg.c
    3
    >>> cfg.d = 4
    >>> cfg['d']
    Traceback (most recent call last):
    ...
    KeyError: 'd'
    >>> cfg.update(**Config({'a':4, 'd':5, 'e':6}))
    >>> cfg.a
    4
    >>> cfg['d']
    Traceback (most recent call last):
    ...
    KeyError: 'd'
    >>> cfg.e
    Traceback (most recent call last):
    ...
    AttributeError: 'Config' object has no attribute 'e'
    '''
    def __init__(
        self, *args, 
        prefix: str = ">>>",
        **kwargs
    ):
        super(Config, self).__init__(*args, **kwargs)
        for name, attr in self.items():
            self.__setattr__(name, attr)
        
        self.prefix = prefix

    def __setitem__(self, key: str, value: Any) -> None:
        super(Config, self).__setitem__(key, value)
        self.__setattr__(key, value)

    def update(self, **kwargs) -> None:
        # Note that, we only update the keys and 
        # correspoding values existed.
        for key, value in kwargs.items():
            if key in self.keys():
                self[key] = value

    def __str__(self) -> str:
        item = "[{name}: {val}] \n"
        infos = f"[{self.__class__.__name__.upper()}] " + self.prefix + "\n"
        for name, val in self.items():
            infos += item.format(name=name, val=val)
        return infos





if __name__ == "__main__":
    import doctest
    doctest.testmod()











